/* File: $Id: device.cc,v 1.20 2004/03/14 10:41:55 morsiani Exp morsiani $ */

/****************************************************************************
 *
 * This module contains definitions for Device class and sub-classes.  Each
 * Device sub-class represents a device type inside the simulation; each
 * object is a single device.  Device class iself defines interface for all
 * devices, and represents the "uninstalled" device.  All devices are
 * created and controlled by a SystemBus object.  Each Device is identified
 * by a (interrupt line, device number) pair.  See external documentation
 * for individual devices capabilities and command/error codes.
 *
 ****************************************************************************/


/****************************************************************************/
/* Inclusion of header files.                                               */
/****************************************************************************/

#include <stdio.h>
#include <string.h>	// kMcC ;)
#include <errno.h>

#include "h/const.h"
#include "h/types.h"
#include "h/blockdev.h"


/****************************************************************************/
/* Inclusion of imported declarations.                                      */
/****************************************************************************/

#include "e/setup.e"
#include "e/blockdev.e"
#include "e/systembus.e"
#include "e/utility.e"


/****************************************************************************/
/* Declarations strictly local to the module.                               */
/****************************************************************************/

// last operation result description
HIDDEN char * opResult[2] =	{	"UNSUCCESSFUL",
								"SUCCESSFUL"
							};


// static buffers definition and costants
#define SMALLBUFSIZE	32
#define STRBUFSIZE 256
HIDDEN char strbuf[STRBUFSIZE];


// common device register definitions
#define STATUS 	0
#define COMMAND 1
#define DATA0	2
#define DATA1	3

// common status conditions
#define READY	1
#define ILOPERR	2
#define BUSY	3


// common command values
#define RESET	0
#define ACK		1

//
// PrinterDevice specific commands, error codes, completion times (in microseconds)
//

// static buffer size
#define PRNTBUFSIZE	128

#define PRNTCHR	2

#define PRNTERR	4

#define PRNTRESETTIME	40
#define PRNTCHRTIME		8

// this means a throughput of 125 KB/s


// TerminalDevice specific definitions, commands, status codes, completion
// times

// static buffer size
#define TERMBUFSIZE	128

// terminal register names
#define RECVSTATUS 	0
#define RECVCOMMAND 1
#define TRANSTATUS	2
#define TRANCOMMAND	3

// specific terminal commands
#define TRANCHR	2
#define RECVCHR	2

// specific terminal status conditions
#define TRANERR 4
#define RECVERR	4
#define RECVD	5
#define TRANSMD 5

// terminal op completion times (in microseconds)
#define TERMRESETTIME	400	
#define TRANCHRTIME		80
#define RECVCHRTIME		80

// this means a throughput of about 12.5 KB/s


// DiskDevice specific commands / status codes 

// static buffer size
#define DISKBUFSIZE	128

// controller reset time (microsecs)
#define DISKRESETTIME	400	

// other performance figures are loaded from file

// controller commands
#define SEEKCYL	2
#define READBLK	3
#define WRITEBLK	4

// specific error codes
#define SEEKERR	4
#define READERR	5
#define WRITERR	6
#define DMAERR	7


// TapeDevice specific commands / status codes

// static buffer size
#define TAPEBUFSIZE	128

// specific commands
#define SKIPBLK	2
// READBLK (3) is the same as for DiskDevice
#define BACKBLK	4

// error codes
#define SKIPERR	4
// READERR (5)  and DMAERR (7) as for DiskDevice too
#define BACKERR	6

// tape op completion times (microsecs)
#define TAPERESETTIME	200
#define REWBLKTIME	60
#define BACKBLKTIME	100
#define READBLKTIME	122
#define SKIPBLKTIME	100

// this means a throughput of 2000 KB/s with 1 Mhz DMA transfer, and
// around 4032 KB/s with 99 Mhz DMA transfer
// It would read a full 1GB cartridge in around 9 minutes (1 Mhz DMA)
// It would rewind from end an 1 GB cartridge in around 2 minutes, and
// fast forward it from start to end (or back) in 3 min 50 secs 

// alternative for it would be 100, 200, 360, 200 that would mean (1GB cart.)
// transfer rate of 1024 -> 1440 KB/s in 17 min -> 13 min, full rewind
// in 3 min 50 secs, fast forward in  7 min


//
// local functions
//

// This function decodes device STATUS field and tells if previous operation
// has been successful or not
HIDDEN const char * isSuccess(unsigned int devType, Word regVal);


/****************************************************************************/
/* Definitions to be exported.                                              */
/****************************************************************************/

// Device class defines the interface to all device types, and represents
// the "uninstalled device" (NULLDEV) itself. Device objects are created and
// controlled by a SystemBus object, but also may be inspected by Watch if
// needed

class Device
{
	public:
	
		// This method creates a Device object with "coordinates" (interrupt
		// line, device number) clears device register, and links it to bus
		Device(SystemBus * busl, unsigned int intl, unsigned int dnum);
	
		virtual ~Device(void);

		// This method returns device type ID
		unsigned int Type(void);
	
		// This method is invoked by SystemBus when the required operation
		// scheduled on the device (using Event objects and queue) should be
		// completed: the default NULLDEV device has nothing to do, but
		// others do
		virtual unsigned int CompleteDevOp(void);

		// This method allows SystemBus to write into device register for
		// device: NULLDEV device register write has no effects, but other
		// devices will start performing required operations if COMMAND
		// register is written with proper codes
		virtual void WriteDevReg(unsigned int regnum, Word data);

		// This method returns the static buffer contained in each device
		// describing the current device status (operation performed, etc.).
		// NULLDEV devices are not operational
		virtual const char * getDevSStr(void);
		
		// This method returns a human-readable expression for completion time of
		// device operation under execution, if one
		virtual const char * getDevCTStr(void);				

		// This method allows to copy inputstr contents inside
		// TerminalDevice receiver buffer: not operational for all other
		// devices (NULLDEV included) and produces a panic message
		virtual void Input(const char * inputstr);

		// This method allows to load/unload tapes inside a TapeDevice. For
		// it, if tFName == NULL or EMPTYSTR, method returns TRUE if a new
		// tape may be loaded, FALSE otherwise; else, if tFName != NULL it
		// tries to load the specified tape file, returning completion
		// status. This method is not operational for other devices (NULLDEV
		// included) and produces a panic message
		virtual Boolean TapeLoad(const char * tFName);

		// This method returns the current value for device register field
		// indexed by regnum
		Word ReadDevReg(unsigned int regnum);

		// This method gets the current operational status for the device,
		// as set by user inside the simulation; a "not-working" device
		// fails all operations requested and reports proper error codes; a
		// NULLDEV always fail
		Boolean getDevNotWorking(void);

		// This method sets the operational status for the device inside the
		// simulation as user wishes; a "not-working" device fails all operations
		// requested and reports proper error codes; a NULLDEV always fail
		Boolean setDevNotWorking(Boolean cond);

		// This method returns old value of statChanged and sets it to
		// FALSE: this way any change into device status may be closely
		// monitored
		Boolean ClearDevStatChange(void);

		
	protected:

		// interrupt line and device number for device 
		unsigned int intL;
		unsigned int devNum;

		// device register structure
		Word reg[DEVREGLEN];

		// device type ID (see h/const.h)
		unsigned int dType;

		// controlling SystemBus object
		SystemBus * bus;

		// completion time TimeStamp for current operation (if one)
		TimeStamp * complTime;

		// device operational status 
		Boolean isWorking;

		// device status change flag
		Boolean statChanged;
};


// This method creates a Device object with "coordinates" (interrupt line,
// device number) clears device register, and links it to bus
Device::Device(SystemBus * busl, unsigned int intl, unsigned int dnum)
{
	unsigned int i;
	
	intL = intl;
	devNum = dnum;
	dType = NULLDEV;
	for (i = 0; i < DEVREGLEN; i++)
		reg[i] = 0UL;
	bus = busl;
	complTime = NULL;
	
	// a NULLDEV never works
	isWorking = FALSE;
	
	// at creation status is surely changed...
	statChanged = TRUE;
}

// No operation for "uninstalled" devices
Device::~Device(void)
{
}


// This method returns device type ID (see h/const.h)
unsigned int Device::Type(void)
{
	return(dType);
}


// This method is invoked by SystemBus when the required operation scheduled
// on the device (using Event objects and queue) should be completed: the
// default NULLDEV device has nothing to do, but others do
unsigned int Device::CompleteDevOp(void)
{
	return(0);
}


// This method allows SystemBus to write into device register for device:
// NULLDEV device register write has no effects, but other devices will
// start performing required operations if COMMAND register is written with
// proper codes
void Device::WriteDevReg(unsigned int dummynum, Word dummydata)
{
}


// This method returns the static buffer contained in each device describing
// the current device status (operation performed, etc.). NULLDEV devices
// are not operational
const char * Device::getDevSStr(void)
{
	return("Not operational");
}


// This method returns the current value for device register field indexed
// by regnum
Word Device::ReadDevReg(unsigned int regnum)
{
	if (regnum < DEVREGLEN)
		return(reg[regnum]);
	else
	{
		Panic("Unknown register index in Device::ReadDevReg()");
		// never returns	
		return(MAXWORDVAL);
	}
}


// This method gets the current operational status for the device, as set by
// user inside the simulation; a "not-working" device fails all operations
// requested and reports proper error codes; a NULLDEV always fail
Boolean Device::getDevNotWorking(void)
{
	return(!isWorking);
}

// This method returns a human-readable expression for completion time of
// device operation under execution, if one
const char * Device::getDevCTStr(void)
{
	if (complTime != NULL)
	{
		sprintf(strbuf, "0x%.8lX.%.8lX", complTime->getHiTS(), complTime->getLoTS());
		return(strbuf);
	}
	else
		return(EMPTYSTR);
}


// This method sets the operational status for the device inside the
// simulation as user wishes; a "not-working" device fails all operations
// requested and reports proper error codes; a NULLDEV always fail
Boolean Device::setDevNotWorking(Boolean cond)
{
	if (dType == NULLDEV)
		isWorking = FALSE;
	else
		isWorking = !cond;
	
	return(!isWorking);
}


// This method returns old value of statChanged and sets it to FALSE: this
// way any change into device status may be closely monitored
Boolean Device::ClearDevStatChange(void)
{
	Boolean old;
	
	old = statChanged;
	statChanged = FALSE;
	return(old);
}


// This method allows to copy inputstr contents inside TerminalDevice
// receiver buffer: not operational for all other devices (NULLDEV included)
// and produces a panic message
void Device::Input(const char * inputstr)
{
	Panic("Input directed to a non-Terminal device in Device::Input()");
}


// This method allows to load/unload tapes inside a TapeDevice. For it, if
// tFName == NULL or EMPTYSTR, method returns TRUE if a new tape may be
// loaded, FALSE otherwise; else, if tFName != NULL it tries to load the
// specified tape file, returning completion status. This method is not
// operational for other devices (NULLDEV included) and produces a panic
// message
Boolean Device::TapeLoad(const char * tFName)
{
	Panic("Tape load request to a non-Tape device in Device::TapeLoad()");
	// no real return
	return(FALSE);
}

			
/****************************************************************************/

// PrinterDevice class allows to emulate parallel character printer
// currently in use (see performance figures shown before). It uses the same
// interface as Device, redefining only a few methods' implementation: refer
// to it for individual methods descriptions. 
// It adds to Device data structure:
// a pointer to SetupInfo object containing printer log file name;
// a static buffer for device operation & status description;
// a FILE structure for log file access.

class PrinterDevice : public Device
{
	public:
		PrinterDevice(SystemBus * busl, SetupInfo * stp, unsigned int intl, unsigned int dnum);
		virtual ~PrinterDevice(void);
		virtual void WriteDevReg(unsigned int regnum, Word data);
		virtual unsigned int CompleteDevOp(void);
		virtual const char * getDevSStr(void);
		
	private:
 		// to be used for getting printer output file name...
		SetupInfo * setup;

		// log file handling
		FILE * prntFile;
		
		// static buffer
		char statStr[PRNTBUFSIZE];
};


//
// See Device class methods description for interface
//

PrinterDevice::PrinterDevice (SystemBus * busl, SetupInfo * stp, unsigned int intl, unsigned int dnum)
							 : Device(busl, intl, dnum)
{
	// adds to a Device object PrinterDevice-specific fields
	dType = PRNTDEV;
	isWorking = TRUE;
	reg[STATUS] = READY;
	sprintf(statStr, "Idle");
	setup = stp;
	
	// tries to open log file
	if ((prntFile = fopen(setup->getDevFileName(intL, devNum), "w")) == NULL)
	{
		sprintf(strbuf, "Cannot open printer %u file : %s", devNum, strerror(errno));
		Panic(strbuf);
	}
	// else file has been open with success
}	


PrinterDevice::~PrinterDevice(void)
{
	// tries to close log file
	if (fclose(prntFile) == EOF)
	{
		sprintf(strbuf, "Cannot close printer file %u : %s", devNum, strerror(errno));
		Panic(strbuf);
	}
}


void PrinterDevice::WriteDevReg(unsigned int regnum, Word data)
{
	// only COMMAND and DATA0 registers are writable, and only when 
	// device is not busy
	
	if (reg[STATUS] != BUSY)
		switch (regnum)
		{
			case COMMAND:
				statChanged = TRUE;
				reg[COMMAND] = data;
				
				// decode operation requested: for each, acknowledges a
				// previous interrupt if pending, sets the device registers,
				// and inserts an Event in SystemBus mantained queue
				
				switch (data)
				{
					case RESET:
						bus->IntAck(intL, devNum);
						complTime = bus->EventReq(intL, devNum, PRNTRESETTIME * setup->getSpeed());
						sprintf(statStr, "Resetting (last op: %s)", isSuccess(dType, reg[STATUS]));
						reg[STATUS] = BUSY;
						break;
				
					case ACK:
						bus->IntAck(intL, devNum);
						sprintf(statStr, "Idle (last op: %s)", isSuccess(dType, reg[STATUS]));
						reg[STATUS] = READY;
						break;
						
					case PRNTCHR:
						bus->IntAck(intL, devNum);
						sprintf(statStr, "Printing char 0x%.2X (last op: %s)", 
							(unsigned char) reg[DATA0], isSuccess(dType, reg[STATUS]));
						complTime = bus->EventReq(intL, devNum, PRNTCHRTIME * setup->getSpeed());
						reg[STATUS] = BUSY;
						break;
				
					default:
						sprintf(statStr, "Unknown command (last op: %s)", isSuccess(dType, reg[STATUS]));
						reg[STATUS] = ILOPERR;
						bus->IntReq(intL, devNum);
						break;
				}
				break;					 

			case DATA0:
				reg[DATA0] = data;
				break;

			default:
				break;
		}
}	


const char * PrinterDevice::getDevSStr(void)
{
	return(statStr);
}


unsigned int PrinterDevice::CompleteDevOp(void)
{

	// checks which operation must be completed: for each, sets device
	// register, performs requested operation and produces an interrupt
	// request
	
	switch (reg[COMMAND])
	{
		case RESET:
			// a reset always works, even if isWorking == FALSE
			sprintf(statStr, "Reset completed : waiting for ACK");
			reg[STATUS] = READY;
			break;
		
		case PRNTCHR:
			if (isWorking)
			{
				// normal operation
				if (fputc((unsigned char) reg[DATA0], prntFile) == EOF)
				{
					sprintf(strbuf, "Error writing printer %u file : %s", devNum, strerror(errno));
					Panic(strbuf);
				}
				// else operation is successful:
				sprintf(statStr, "Printed char 0x%.2X : waiting for ACK", (unsigned char) reg[DATA0]);
				reg[STATUS] = READY;
			}
			else
			{
				// no operation & error simulation
				sprintf(statStr, "Error printing char 0x%.2X : waiting for ACK", (unsigned char) reg[DATA0]);
				reg[STATUS] = PRNTERR;
			}
			break;

		default:
			Panic("Unknown operation in PrinterDevice::CompleteDevOp()");
			break;
	}
	statChanged = TRUE;
	complTime = NULL;
	bus->IntReq(intL, devNum);
	return(STATUS);
}


/****************************************************************************/


// TerminalDevice class allows to emulate serial "dumb" terminal (see
// performance figures shown before). TerminalDevice may be split up into
// two independent sub-devices: "transmitter" and "receiver".
// It uses the same interface as Device, redefining only a few methods'
// implementation: refer to it for individual methods descriptions.
// It adds to Device data structure:
// a pointer to SetupInfo object containing terminal log file name;
// a static buffer for device operation & status description;
// a FILE structure for log file access;
// some structures for handling terminal transmitter and receiver.

class TerminalDevice : public Device
{
	public:
		TerminalDevice(SystemBus * busl, SetupInfo * stp, unsigned int intl, unsigned int dnum);
		virtual ~TerminalDevice(void);
		virtual void WriteDevReg(unsigned int regnum, Word data);
		virtual unsigned int CompleteDevOp(void);
		virtual const char * getDevSStr(void);
		virtual const char * getDevCTStr(void);
		virtual void Input(const char * inputstr);
		
	private:
		// for log file name
		SetupInfo * setup;
		// for log file handling
		FILE * termFile;
		
		// receiver buffer and pointer to first character to receive from it
		char * recvBuf;
		unsigned int recvBp;
		
		// static buffer for receiver
		char recvStatStr[TERMBUFSIZE];
		
		// static buffer for transmitter
		char tranStatStr[TERMBUFSIZE];
		
		// completion time TimeStamp for current receiver operation (if one)
		TimeStamp * recvCTime;

		// completion time TimeStamp for current transmitter operation (if one)
		TimeStamp * tranCTime;
		
		// receiver operation pending flag
		Boolean recvIntPend;		
		
		// transmitter operation pending flag
		Boolean tranIntPend;
};


//
// See Device class methods description for interface
//

TerminalDevice::TerminalDevice (SystemBus * busl, SetupInfo * stp, unsigned int intl, unsigned int dnum)
							 : Device(busl, intl, dnum)
{
	// adds to a Device object TerminalDevice specific fields
	dType = TERMDEV;
	isWorking = TRUE;
	recvBuf = NULL;
	recvBp = 0;
	reg[RECVSTATUS] = READY;
	reg[TRANSTATUS] = READY;
	sprintf(recvStatStr, "Idle");
	sprintf(tranStatStr, "Idle");
	recvCTime = NULL;
	tranCTime = NULL;
	recvIntPend = FALSE;
	tranIntPend = FALSE;
	setup = stp;
	
	// tries to open log file
	if ((termFile = fopen(setup->getDevFileName(intL, devNum), "w")) == NULL)
	{
		sprintf(strbuf, "Cannot open terminal %u file : %s", devNum, strerror(errno));
		Panic(strbuf);
	}
	// else file has been open with success: set it to no buffering for quick 
	// terminal screen update
	setvbuf(termFile, (char *) NULL, _IONBF, 0);
}	


TerminalDevice::~TerminalDevice(void)
{
	if (fclose(termFile) == EOF)
	{
		sprintf(strbuf, "Cannot close terminal file %u : %s", devNum, strerror(errno));
		Panic(strbuf);
	}
}


void TerminalDevice::WriteDevReg(unsigned int regnum, Word data)
{
	// only COMMAND registers are writable, and only when device is not busy
	// format is NNNN NNNN CHAR COMM
	
	switch (regnum)
	{
		case RECVCOMMAND:
		
			// decode operation requested: for each, acknowledges a
			// previous interrupt if pending, sets the device registers,
			// and inserts an Event in SystemBus mantained queue
			                                
			if (reg[RECVSTATUS] != BUSY)
			{
				statChanged = TRUE;
				reg[RECVCOMMAND] = data;
				
				// to extract command
				switch (data & BYTEMASK)
				{
					case RESET:
						if (!tranIntPend)
							bus->IntAck(intL, devNum);
						recvIntPend = FALSE;
						recvCTime = bus->EventReq(intL, devNum, TERMRESETTIME * setup->getSpeed());
						sprintf(recvStatStr, "Resetting (last op: %s)", 
							isSuccess(dType, reg[RECVSTATUS] & BYTEMASK));
						reg[RECVSTATUS] = BUSY;
						break;
						
					case ACK:
						if (!tranIntPend)
							bus->IntAck(intL, devNum);
						recvIntPend = FALSE;
						sprintf(recvStatStr, "Idle (last op: %s)",\
						isSuccess(dType, reg[RECVSTATUS] & BYTEMASK));
						reg[RECVSTATUS] = READY;
						break;
					
					case RECVCHR:
						if (!tranIntPend)
							bus->IntAck(intL, devNum);
						recvIntPend = FALSE;
						sprintf(recvStatStr, "Receiving (last op: %s)", \
							isSuccess(dType, reg[RECVSTATUS] & BYTEMASK));
						recvCTime = bus->EventReq(intL, devNum, RECVCHRTIME * setup->getSpeed());
						reg[RECVSTATUS] = BUSY;
						break;
						
					default:
						sprintf(recvStatStr, "Unknown command (last op: %s)", \
							isSuccess(dType, reg[RECVSTATUS] & BYTEMASK));
						reg[RECVSTATUS] = ILOPERR;
						bus->IntReq(intL, devNum);
						recvIntPend = TRUE;
						break;
				}	
			}
			// else no change
			break;
		
		case TRANCOMMAND:
		
			// decode operation requested: for each, acknowledges a
			// previous interrupt if pending, sets the device registers,
			// and inserts an Event in SystemBus mantained queue
			                                
			if (reg[TRANSTATUS] != BUSY)
			{
				reg[TRANCOMMAND] = data;
				statChanged = TRUE;
					
				// to extract command
				switch (data & BYTEMASK)
				{
					case RESET:
						if (!recvIntPend)
							bus->IntAck(intL, devNum);
						tranIntPend = FALSE;
						tranCTime = bus->EventReq(intL, devNum, TERMRESETTIME * setup->getSpeed());
						sprintf(tranStatStr, "Resetting (last op: %s)",\
							isSuccess(dType, reg[TRANSTATUS] & BYTEMASK));
						reg[TRANSTATUS] = BUSY;
						break;
						
					case ACK:
						if (!recvIntPend)
							bus->IntAck(intL, devNum);
						tranIntPend = FALSE; 
						sprintf(tranStatStr, "Idle (last op: %s)",\
							isSuccess(dType, reg[TRANSTATUS] & BYTEMASK));
						reg[TRANSTATUS] = READY;
						break;
					
					case TRANCHR:
						if (!recvIntPend)
							bus->IntAck(intL, devNum);
						tranIntPend = FALSE;
						sprintf(tranStatStr, "Transm. char 0x%.2X (last op: %s)", \
							(unsigned char) ((data >> BYTELEN) & BYTEMASK),\
							isSuccess(dType, reg[TRANSTATUS] & BYTEMASK));
							
						tranCTime = bus->EventReq(intL, devNum, TRANCHRTIME * setup->getSpeed());
						reg[TRANSTATUS] = BUSY;
						break;		
						
					default:
						sprintf(tranStatStr, "Unknown command (last op: %s)",\
							isSuccess(dType, reg[TRANSTATUS] & BYTEMASK));
						reg[TRANSTATUS] = ILOPERR;
						bus->IntReq(intL, devNum);
						tranIntPend = TRUE;
						break;
				}
			}
			// else no change
			break;
			
		case RECVSTATUS:
		case TRANSTATUS:
		default:
			// no change
			break;
	}
}	


const char * TerminalDevice::getDevSStr(void)
{
	sprintf(strbuf, "%s\n%s", recvStatStr, tranStatStr); 
	return(strbuf);
}


const char * TerminalDevice::getDevCTStr(void)
{
	char tmpbuf[SMALLBUFSIZE];

	// There two completion times must be returned, if needed: they are 
	// split up by a newline
	 	 
	if (recvCTime != NULL)
		sprintf(tmpbuf, "0x%.8lX.%.8lX", recvCTime->getHiTS(), recvCTime->getLoTS());
	else
		strcpy(tmpbuf, EMPTYSTR);
	if (tranCTime != NULL)
		sprintf(strbuf, "%s\n0x%.8lX.%.8lX", tmpbuf, tranCTime->getHiTS(), tranCTime->getLoTS());
	else
		sprintf(strbuf, "%s\n", tmpbuf);
	
	return(strbuf);			
}


unsigned int TerminalDevice::CompleteDevOp(void)
{
	// only one sub-device should complete its op: which one?
	Boolean doRecv;
	unsigned int devMod;
	
	// determines which operation must be completed
	if (reg[RECVSTATUS] == BUSY && reg[TRANSTATUS] == BUSY)
	{
		// both sub-devices are working, so tie breaking depends on 
		// timestamps: lower is first. If they are equal, this don't matter
		// because there will be another CompleteDevOp following, and one 
		// sub-device will have already completed its op or posponed it (recv)
		if (recvCTime->LessEq(tranCTime))
			doRecv = TRUE;
		else
			doRecv = FALSE;
	}
	else
		// surely one or other sub-device is idle
		doRecv = (reg[RECVSTATUS] == BUSY);
	
	if (doRecv)				
	{
		// recv sub-device operation completion
		switch (reg[RECVCOMMAND])
		{
			// checks which operation must be completed: for each, sets device
			// register, performs requested operation and produces an interrupt
			// request				
			
			case RESET:
				// a reset always works, even if isWorking == FALSE
				sprintf(recvStatStr, "Reset completed : waiting for ACK");
				reg[RECVSTATUS] = READY;
				recvIntPend = TRUE;
				recvCTime = NULL;
				bus->IntReq(intL, devNum);
				break;
		
			case RECVCHR:
				if (recvBuf == NULL || recvBuf[recvBp] == EOS)
					// no char in input: wait another receiver cycle
					recvCTime = bus->EventReq(intL, devNum, RECVCHRTIME * setup->getSpeed());
				else
				{	
					// buffer is not empty
					if (isWorking)
					{
						sprintf(recvStatStr, "Received char 0x%.2X : waiting for ACK", recvBuf[recvBp]);
						reg[RECVSTATUS] = (((Word) recvBuf[recvBp]) << BYTELEN) | RECVD;
						recvBp++;
					}
					else
					{
						// no operation & error simulation
						sprintf(recvStatStr, "Error receiving char : waiting for ACK");
						reg[RECVSTATUS] = RECVERR;
					}
					// interrupt request
					recvIntPend = TRUE;
					recvCTime = NULL;
					bus->IntReq(intL, devNum);
				}	
				break;

		default:
			Panic("Unknown operation in TerminalDevice::CompleteDevOp()");
			break;
		}
		// even if there is no char to receive, the RECVSTATUS register is
		// "rewritten" by the receiver and CTime changed
		devMod = RECVSTATUS;

	}
	else
	{
		// transmitter sub-device operation completion
		switch (reg[TRANCOMMAND] & BYTEMASK)
		{
			// checks which operation must be completed: for each, sets device
			// register, performs requested operation and produces an interrupt
			// request		

			case RESET:
				// a reset always works, even if isWorking == FALSE
				sprintf(tranStatStr, "Reset completed : waiting for ACK");
				reg[TRANSTATUS] = READY;
				break;
		
			case TRANCHR:
				if (isWorking)
				{
					if (fputc((unsigned char) ((reg[TRANCOMMAND] >> BYTELEN) & BYTEMASK), termFile) == EOF)
					{
						sprintf(strbuf, "Error writing terminal %u file : %s", devNum, strerror(errno));
						Panic(strbuf);
					}
					// else operation is successful:
					sprintf(tranStatStr, "Transm. char 0x%.2lX : waiting for ACK",\
						(reg[TRANCOMMAND] >> BYTELEN) & BYTEMASK);
					reg[TRANSTATUS] = (reg[TRANCOMMAND] & (BYTEMASK << BYTELEN)) | TRANSMD;
					
				}
				else
				{
					// no operation & error simulation
					sprintf(tranStatStr, "Error transm. char 0x%.2lX : waiting for ACK",\
						(reg[TRANCOMMAND] >> BYTELEN) & BYTEMASK);
					reg[TRANSTATUS] = (reg[TRANCOMMAND] & (BYTEMASK << BYTELEN)) | TRANERR;
				}
				break;

		default:
			Panic("Unknown operation in TerminalDevice::CompleteDevOp()");
			break;
		}
		// interrupt generation 
		bus->IntReq(intL, devNum);
		tranIntPend = TRUE;
		tranCTime = NULL;
		devMod = TRANSTATUS;
	}
	statChanged = TRUE;
	return(devMod);
	
}


void TerminalDevice::Input(const char * inputstr)
{
	char * strp;
	
	if (recvBuf != NULL && recvBuf[recvBp] == EOS)
	{
		// buffer exausted: delete it
		delete recvBuf;
		recvBuf = NULL;
	}
	
	if (recvBuf == NULL)
	{
		// simply copies inputstr into it, adding a trailing '\n'
		recvBuf = new char [strlen(inputstr) + 2];
		sprintf(recvBuf, "%s\n", inputstr);
	}
	else
	{
		// copies the unreceived buffer part plus the new input into
		// a new buffer
		strp = new char [(strlen(recvBuf) - recvBp) + strlen(inputstr) + 2];
		sprintf(strp, "%s%s\n",&(recvBuf[recvBp]), inputstr);
		delete recvBuf;
		recvBuf = strp;
	}
	recvBp = 0;
	statChanged = TRUE;

	// writes input to log file	
	if (fprintf(termFile, "%s\n", inputstr) < 0)
	{
		sprintf(strbuf, "Error writing terminal %u file : %s", devNum, strerror(errno));
		Panic(strbuf);
	}
}  	


/****************************************************************************/


// DiskDevice class allows to emulate a disk drive: each 512 byte sector it
// contains is identified by (cyl, head, sect) set of disk coordinates;
// (geometry and performance figures are loaded from disk image file). 
// Operations on sectors (R/W) require previous seek on the desired cylinder.
// It also contains a sector buffer of one sector to speed up operations.
//
// It uses the same interface as Device, redefining only a few methods'
// implementation: refer to it for individual methods descriptions.
//
// It adds to Device data structure:
// a pointer to SetupInfo object containing disk image file name;
// a static buffer for device operation & status description;
// a FILE structure for disk image file access;
// a set of disk parameters (read from disk image file header);
// a Block object for file handling;
// some items for performance computation.

class DiskDevice : public Device
{
	public:
		DiskDevice(SystemBus * busl, SetupInfo * stp, unsigned int intl, unsigned int dnum);
		virtual ~DiskDevice(void);
		virtual void WriteDevReg(unsigned int regnum, Word data);
		virtual unsigned int CompleteDevOp(void);
		virtual const char * getDevSStr(void);
		
	private:
	
		// to get disk image file name
		SetupInfo * setup;
		
		// to handle it
		FILE * diskFile;
		
		// static buffer
		char statStr[DISKBUFSIZE];
		
		// sector buffer and coordinates on disk (cyl, head, sect)
		Block * diskBuf;
		unsigned int cylBuf, headBuf, sectBuf;
		
		// start of disk image inside file (after header)
		SWord diskOfs;
		
		// disk performance parameters
		DriveParams * diskP;
		
		// sector underhead time in ticks
		Word sectTicks;

		// current cylinder
		unsigned int currCyl;
};


//
// See Device class methods description for interface
//
 
DiskDevice::DiskDevice (SystemBus * busl, SetupInfo * stp, unsigned int intl, unsigned int dnum)
							 : Device(busl, intl, dnum)
{
	// adds to a Device object DiskDevice-specific fields
	dType = DISKDEV;
	isWorking = TRUE;
	reg[STATUS] = READY;
	sprintf(statStr, "Idle");
	diskBuf = new Block();
	
	setup = stp;
	
	// tries to access disk image file 
	if ((diskFile = fopen(setup->getDevFileName(intL, devNum), "r+")) == NULL)
	{
		sprintf(strbuf, "Cannot open disk %u file : %s", devNum, strerror(errno));
		Panic(strbuf);
	}
	
	// else file has been open with success: tests if it is a valid disk file
	diskP = new DriveParams(diskFile, &diskOfs);	
	
	if (diskOfs == 0)
	{
		// file is not a valid disk file
		sprintf(strbuf, "Cannot open disk %u file : invalid/corrupted file", devNum);
		Panic(strbuf);
	}
	
	// else all is OK

	// DATA1 format == drive geometry: CYL CYL HEAD SECT
	reg[DATA1] = (diskP->getCylNum() << HWORDLEN) | (diskP->getHeadNum() << BYTELEN) | diskP->getSectNum(); 

	currCyl = 0;
	sectTicks = ((diskP->getRotTime() * setup->getSpeed()) / diskP->getSectNum());
	cylBuf = headBuf = sectBuf = MAXWORDVAL;
}   


DiskDevice::~DiskDevice(void)
{
	delete diskBuf;
	delete diskP;

	if (fclose(diskFile) == EOF)
	{
		sprintf(strbuf, "Cannot close disk file %u : %s", devNum, strerror(errno));
		Panic(strbuf);
	}
	
}


void DiskDevice::WriteDevReg(unsigned int regnum, Word data)
{
	// only COMMAND and DATA0 registers are writable, and only when 
	// device is not busy
	Word timeOfs;
	unsigned int cyl, head, sect, currSect;
	
	if (reg[STATUS] != BUSY)
		switch (regnum)
		{
			case COMMAND:
				statChanged = TRUE;
				reg[COMMAND] = data;
				switch (data & BYTEMASK)
				{

					// decode operation requested: for each, acknowledges a
					// previous interrupt if pending, sets the device registers,
					// and inserts an Event in SystemBus mantained queue				

					case RESET:
						bus->IntAck(intL, devNum);
						// controller reset & cylinder recalibration
						timeOfs = (DISKRESETTIME + (diskP->getSeekTime() * currCyl)) * setup->getSpeed();
						complTime = bus->EventReq(intL, devNum, timeOfs);
						sprintf(statStr, "Resetting (last op: %s)", isSuccess(dType, reg[STATUS]));
						reg[STATUS] = BUSY;
						break;
				
					case ACK:
						bus->IntAck(intL, devNum);
						sprintf(statStr, "Idle (last op: %s)", isSuccess(dType, reg[STATUS]));
						reg[STATUS] = READY;
						break;
						
					case SEEKCYL:
						bus->IntAck(intL, devNum);
						cyl = (data >> BYTELEN) & IMMMASK;
						if (cyl < diskP->getCylNum())
						{  
							bus->IntAck(intL, devNum);
							sprintf(statStr, "Seeking Cyl 0x%.4X (last op: %s)", cyl,\
								isSuccess(dType, reg[STATUS]));
							// compute movement offset 
							if (cyl < currCyl)
								cyl = currCyl - cyl;
							else
								cyl = cyl - currCyl;
							complTime = bus->EventReq(intL, devNum, (diskP->getSeekTime() * cyl * setup->getSpeed()) + 1);
							reg[STATUS] = BUSY;
						}
						else
						{
							// cyl out of range 
							sprintf(statStr, "Cyl 0x%.4X out of range : waiting for ACK", cyl);
							reg[STATUS] = SEEKERR;
							bus->IntReq(intL, devNum);
						}
						break;
				
					case READBLK:
						bus->IntAck(intL, devNum);
						// computes target coordinates
						head = (data >> HWORDLEN) & BYTEMASK;
						sect = (data >> BYTELEN) & BYTEMASK;
						if (head < diskP->getHeadNum() && sect < diskP->getSectNum())
						{
							sprintf(statStr, "Reading C/H/S 0x%.4X/0x%.2X/0x%.2X (last op: %s)",\
								currCyl, head, sect, isSuccess(dType, reg[STATUS]));
							if (currCyl == cylBuf && head == headBuf && sect == sectBuf)
								// sector is already in disk buffer
								timeOfs = DMATICKS;
							else
							{
								// invalidate current buffer
								cylBuf = headBuf = sectBuf = MAXWORDVAL;
								
								// compute op completion time
								
								// use only TodLO for easier computation
								currSect = (bus->getToDLO() / sectTicks) % diskP->getSectNum();
								
								// remaining time for current sector
								timeOfs = bus->getToDLO() % sectTicks;
								
								// compute sector offset
								if (sect > currSect)
									sect = (sect - currSect) - 1;
								else
									sect = (diskP->getSectNum() - 1) - (currSect - sect); 
								
								// completion time is = current sect rem. time + 
								//   sectors-in-between time + sector data read +
								// DMA transfer time
								timeOfs += (sectTicks * sect) + ((sectTicks * diskP->getDataSect()) / 100) + DMATICKS; 
							}
							complTime = bus->EventReq(intL, devNum, timeOfs);
							reg[STATUS] = BUSY;
						}
						else
						{
							// head/sector out of range 
							sprintf(statStr, "Head/sect 0x%.2X/0x%.2X out of range : waiting for ACK", head, sect);
							reg[STATUS] = READERR;
							bus->IntReq(intL, devNum);
						}	
						break;
						
					case WRITEBLK:
						bus->IntAck(intL, devNum);
						// computes target coordinates
						head = (data >> HWORDLEN) & BYTEMASK;
						sect = (data >> BYTELEN) & BYTEMASK;
						if (head < diskP->getHeadNum() && sect < diskP->getSectNum())
						{
							sprintf(statStr, "Writing C/H/S 0x%.4X/0x%.2X/0x%.2X (last op: %s)",\
								currCyl, head, sect, isSuccess(dType, reg[STATUS]));
							// DMA transfer from memory
							if (bus->DMATransfer(diskBuf, reg[DATA0], FALSE))
							{
								// DMA transfer error: invalidate current buffer
								cylBuf = headBuf = sectBuf = MAXWORDVAL;
								timeOfs = DMATICKS;
							}
							else
							{
								// disk sector in buffer from memory
								cylBuf = currCyl;
								headBuf = head;
								sectBuf = sect;	
									
								// compute op completion time
									
								// use only TodLO for easier computation
								// disk spins during DMA transfer
								currSect = ((bus->getToDLO() + DMATICKS) / sectTicks) % diskP->getSectNum();
								
								// remaining time for DMA + current sector
								timeOfs = DMATICKS + ((bus->getToDLO() + DMATICKS) % sectTicks);
								
								// compute sector offset
								if (sect > currSect)
									sect = (sect - currSect) - 1;
								else
									sect = (diskP->getSectNum() - 1) - (currSect - sect); 
							
								// completion time is = DMA time + current sect rem. time + 
								//   sectors-in-between time + sector data write
								timeOfs += (sectTicks * sect) + ((sectTicks * diskP->getDataSect()) / 100);
							}
							complTime = bus->EventReq(intL, devNum, timeOfs);
							reg[STATUS] = BUSY;
						}
						else
						{
							// head/sector out of range 
							sprintf(statStr, "Head/sect 0x%.2X/0x%.2X out of range : waiting for ACK", head, sect);
							reg[STATUS] = WRITERR;
							bus->IntReq(intL, devNum);
						}	
					break;
					
					default:
						sprintf(statStr, "Unknown command (last op: %s)", isSuccess(dType, reg[STATUS]));
						reg[STATUS] = ILOPERR;
						bus->IntReq(intL, devNum);
						break;
				}
				break;					 

			case DATA0:
				// physical address for R/W buffer in memory
				reg[DATA0] = data;
				break;

			default:
				break;
		}
}	


const char * DiskDevice::getDevSStr(void)
{
	return(statStr);
}


unsigned int DiskDevice::CompleteDevOp(void)
{
	// for file access
	SWord blkOfs;
	unsigned int head, sect;
	
	switch (reg[COMMAND] & BYTEMASK)
	{
		// checks which operation must be completed: for each, sets device
		// register, performs requested operation and produces an interrupt
		// request

		case RESET:
			// a reset always works, even if isWorking == FALSE
			// it invalidates the sector buffer
			sprintf(statStr, "Reset completed : waiting for ACK");
			reg[STATUS] = READY;
			cylBuf = headBuf = sectBuf = MAXWORDVAL;
			break;
		
		case SEEKCYL:
			if (isWorking)
			{
				currCyl = (reg[COMMAND] >> BYTELEN) & IMMMASK;
				sprintf(statStr, "Cyl 0x%.4X reached : waiting for ACK", currCyl);
				reg[STATUS] = READY;
			}
			else
			{
				// error simulation: currCyl is between seek start & end  
				currCyl = (((reg[COMMAND] >> BYTELEN) & IMMMASK) + currCyl) / 2;
				sprintf(statStr, "Cyl 0x%.4X seek error : waiting for ACK", currCyl);
				reg[STATUS] = SEEKERR;
			}
			break;
		
		case READBLK:
			// locates target coordinates
			head = (reg[COMMAND] >> HWORDLEN) & BYTEMASK;
			sect = (reg[COMMAND] >> BYTELEN) & BYTEMASK;
			if (isWorking)
			{
				blkOfs = (diskOfs + ((currCyl * diskP->getHeadNum() * diskP->getSectNum()) + (head * diskP->getSectNum()) + sect) * BLOCKSIZE) * WORDLEN;
				if (cylBuf != MAXWORDVAL || !diskBuf->ReadBlock(diskFile, blkOfs))
				//	wanted sector is already in buffer or has been read correctly
				{
					cylBuf = currCyl;
					headBuf = head;
					sectBuf = sect;
					if (bus->DMATransfer(diskBuf, reg[DATA0], TRUE))
					{
						// DMA transfer error
						reg[STATUS] = DMAERR;
						sprintf(statStr, "DMA error reading C/H/S 0x%.4X/0x%.2X/0x%.2X : waiting for ACK",\
								currCyl, head, sect);
					}
					else
					{
						// all ok
						sprintf(statStr, "C/H/S 0x%.4X/0x%.2X/0x%.2X block read: waiting for ACK",\
								currCyl, head, sect);
						reg[STATUS] = READY;
					}
				}
				else 
				{
					// ReadBlock() has failed for sure
					sprintf(strbuf, "Unable to read disk %u file : invalid/corrupted file", devNum);
					Panic(strbuf);
				}
			}
			else
			{
				// error simulation
				sprintf(statStr, "Error reading C/H/S 0x%.4X/0x%.2X/0x%.2X : waiting for ACK",\
								currCyl, head, sect);
				// buffer invalidation 
				cylBuf = headBuf = sectBuf = MAXWORDVAL;
				reg[STATUS] = READERR;
			}
			break;	
		
		case WRITEBLK:
			// locates target coordinates
			head = (reg[COMMAND] >> HWORDLEN) & BYTEMASK;
			sect = (reg[COMMAND] >> BYTELEN) & BYTEMASK;
			if (isWorking)
			{
				blkOfs = (diskOfs + ((currCyl * diskP->getHeadNum() * diskP->getSectNum()) + (head * diskP->getSectNum()) + sect) * BLOCKSIZE) * WORDLEN;
				if (diskBuf->WriteBlock(diskFile, blkOfs))
				{
					// error writing block to disk file
					sprintf(strbuf, "Unable to write disk %u file : invalid/corrupted file", devNum);
					Panic(strbuf);				
				}
				// else all is ok: buffer is still valid
				sprintf(statStr, "C/H/S 0x%.4X/0x%.2X/0x%.2X block written : waiting for ACK",\
								currCyl, head, sect);
				reg[STATUS] = READY;
			}
			else
			{
				// error simulation & buffer invalidation 
				cylBuf = headBuf = sectBuf = MAXWORDVAL;
				sprintf(statStr, "Error writing C/H/S 0x%.4X/0x%.2X/0x%.2X : waiting for ACK",\
								currCyl, head, sect);
				reg[STATUS] = WRITERR;
			}
			break;
			
		default:
			Panic("Unknown operation in DiskDevice::CompleteDevOp()");
			break;
	}
	statChanged = TRUE;
	complTime = NULL;
	bus->IntReq(intL, devNum);
	return(STATUS);
}


/****************************************************************************/


// TapeDevice class allows to emulate removable cartridge tape drives. 
// Individual tapes may be loaded and unloaded, rewound and read 
// (see performance figures shown before). TapeDevice uses the same
// interface as Device, redefining only a few methods' implementation: refer
// to it for individual methods descriptions.
// It adds to Device data structure:
// a pointer to SetupInfo object containing tape cartridge log file name;
// a static buffer for device operation & status description;
// a FILE structure for log file access;
// a Block object for file handling.

class TapeDevice : public Device
{
	public:
		TapeDevice(SystemBus * busl, SetupInfo * stp, unsigned int intl, unsigned int dnum);
		virtual ~TapeDevice(void);
		virtual void WriteDevReg(unsigned int regnum, Word data);
		virtual unsigned int CompleteDevOp(void);
		virtual const char * getDevSStr(void);
		virtual Boolean TapeLoad(const char * tFName);

	private:
		// to get tape cartridge file name
		SetupInfo * setup;
		
		// to access tape image file
		FILE * tapeFile;
		char * tapeFName;
		
		// to read tape blocks and know current position (starts with block 0)
		Block * tapeBlk;
		unsigned int tapeBp;
		
		// loaded cartridge flag
		Boolean tapeLoaded;
		
		// static buffer
		char statStr[TAPEBUFSIZE];
};


//
// See Device class methods description for interface
//

TapeDevice::TapeDevice (SystemBus * busl, SetupInfo * stp, unsigned int intl, unsigned int dnum)
							 : Device(busl, intl, dnum)
{
	// adds to a Device object TapeDevice-specific fields
	setup = stp;
	dType = TAPEDEV;
	isWorking = TRUE;
	tapeFName = NULL;
	tapeBlk = new Block();
	tapeBp = 0;
	tapeLoaded = FALSE;
	reg[STATUS] = READY;
	reg[DATA1] = TAPEEOT;
	
	sprintf(statStr, "Idle");
	
	// return value is ignored: a tape will be loaded if its setup file
	// name exists and file is of correct type
	TapeLoad(setup->getDevFileName(intL, devNum));

}	

TapeDevice::~TapeDevice(void)
{
	delete tapeBlk;
	
	if (tapeLoaded)
	{
		delete tapeFName;
	
		if (fclose(tapeFile) == EOF)
		{
			sprintf(strbuf, "Cannot close tape file %u : %s", devNum, strerror(errno));
			Panic(strbuf);
		}
	}
}

// If tFName == NULL or EMPTYSTR, method returns TRUE if 
// a new tape may be loaded,  FALSE otherwise; else, if tFName != NULL
// it tries to load the specified tape file, returning completion status 
Boolean TapeDevice::TapeLoad(const char * tFName)
{
	Word tapeid = 0;
	
	if (tFName == NULL || strlen(tFName) == 0)
		// only a request for tape loading availability
		return(!tapeLoaded || (reg[STATUS] == READY && reg[DATA1] == TAPESTART));
	else
		if (tapeLoaded && (reg[STATUS] != READY || reg[DATA1] != TAPESTART))
			// a new tape cannot be loaded until the current tape is not 
			// rewound: a safeguard to avoid tape ejection during I/O operations
			return(FALSE);
		else
		{
			if (tapeLoaded)
			{
				// a tape is currently loaded
				delete tapeFName;
				if (fclose(tapeFile) == EOF)
				{
					sprintf(strbuf, "Cannot close tape file %u : %s", devNum, strerror(errno));
					Panic(strbuf);
				}
				// else all is OK
			}
			tapeFName = new char [strlen(tFName) + 1];
			strcpy(tapeFName, tFName);
			if ((tapeFile = fopen(tapeFName, "r")) == NULL ||\
				fread((void *) &tapeid, WORDLEN, 1, tapeFile) != 1 || \
				tapeid != TAPEFILEID)
			{
				sprintf(strbuf, "Cannot open tape %u file : invalid/corrupted file", devNum);
				Panic(strbuf);
			}
			// else file opening is OK: setting tape parameters
			reg[DATA1] = TAPESTART;
			tapeLoaded = TRUE;
			tapeBp = 0;
			statChanged = TRUE;
			return(TRUE);
		}
}
		

void TapeDevice::WriteDevReg(unsigned int regnum, Word data)
{
	// only COMMAND and DATA0 registers are writable, and only when 
	// device is not busy and a tape is loaded
	
	if (reg[STATUS] != BUSY && tapeLoaded)
		switch (regnum)
		{


			case COMMAND:

				// decode operation requested: for each, acknowledges a
				// previous interrupt if pending, sets the device registers,
				// and inserts an Event in SystemBus mantained queue
				
				statChanged = TRUE;
				reg[COMMAND] = data;
				switch (data)
				{
					// it rewinds the tape too
					case RESET:
						bus->IntAck(intL, devNum);
						complTime = bus->EventReq(intL, devNum, (TAPERESETTIME + (REWBLKTIME * tapeBp)) * setup->getSpeed());
						sprintf(statStr, "Rewinding the tape (last op: %s)", isSuccess(dType, reg[STATUS]));
						reg[STATUS] = BUSY;
						break;
				
					case ACK:
						bus->IntAck(intL, devNum);
						sprintf(statStr, "Idle (last op: %s)", isSuccess(dType, reg[STATUS]));
						reg[STATUS] = READY;
						break;
						
					case SKIPBLK:
						bus->IntAck(intL, devNum);
						if (reg[DATA1] != TAPEEOT)
						{
							sprintf(statStr, "Skipping block %u (last op: %s)", tapeBp, isSuccess(dType, reg[STATUS]));
							complTime = bus->EventReq(intL, devNum, SKIPBLKTIME * setup->getSpeed());
							reg[STATUS] = BUSY;
						}
						else
						{
							sprintf(statStr, "Cannot skip beyond EOT : waiting for ACK");
							reg[STATUS] = SKIPERR;
							bus->IntReq(intL, devNum);
						}
						break;
						
					case READBLK:
						bus->IntAck(intL, devNum);
						if (reg[DATA1] != TAPEEOT)
						{
							sprintf(statStr, "Reading block %u (last op: %s)", tapeBp, isSuccess(dType, reg[STATUS]));
							complTime = bus->EventReq(intL, devNum, READBLKTIME * setup->getSpeed() + DMATICKS);
							reg[STATUS] = BUSY;
						}
						else 
						{
							sprintf(statStr, "Cannot read beyond EOT : waiting for ACK");
							reg[STATUS] = READERR;
							bus->IntReq(intL, devNum);
						}
						break;	
						
					case BACKBLK:
						bus->IntAck(intL, devNum);
						// overkill test
						if (reg[DATA1] != TAPESTART && tapeBp != 0)
						{
							sprintf(statStr, "Rewinding to block %u (last op: %s)", tapeBp - 1, isSuccess(dType, reg[STATUS]));
							complTime = bus->EventReq(intL, devNum, REWBLKTIME * setup->getSpeed());
							reg[STATUS] = BUSY;
						}
						else
						{
							sprintf(statStr, "Cannot rewind beyond tape start : waiting for ACK");
							reg[STATUS] = BACKERR;
							bus->IntReq(intL, devNum);
						}
						break;
						
					default:
						sprintf(statStr, "Unknown command (last op: %s)", isSuccess(dType, reg[STATUS]));
						reg[STATUS] = ILOPERR;
						bus->IntReq(intL, devNum);
						break;
				}
				break;					 

			case DATA0:
				// sets physical address for buffer in memory
				reg[DATA0] = data;
				break;

			default:
				break;
		}
}	

const char * TapeDevice::getDevSStr(void)
{
	return(statStr);
}


unsigned int TapeDevice::CompleteDevOp(void)
{
	// checks which operation must be completed
	switch (reg[COMMAND])
	{

		// checks which operation must be completed: for each, sets device
		// register, performs requested operation and produces an interrupt
		// request
	
		case RESET:
			// a reset always works, even if isWorking == FALSE
			sprintf(statStr, "Rewind completed : waiting for ACK");
			reg[STATUS] = READY;
			reg[DATA1] = TAPESTART;
			tapeBp = 0;
			break;
		
		case SKIPBLK:
			// a SKIPBLK is always successful (isWorking status does not matter)
			if (tapeBlk->ReadBlock(tapeFile, (tapeBp * BLOCKSIZE * WORDLEN) + ((tapeBp + 1) * WORDLEN)) || \
				fread((void *) &(reg[DATA1]), WORDLEN, 1, tapeFile) != 1 || reg[DATA1] > TAPEEOB)
			{
				sprintf(strbuf, "Error reading tape %u file : %s", devNum, strerror(errno));
				Panic(strbuf);
			}
			// else operation is successful: 
			reg[STATUS] = READY;
			sprintf(statStr, "Block %u skipped : waiting for ACK", tapeBp);
			tapeBp++;
			break;
					
		case READBLK:
			if (tapeBlk->ReadBlock(tapeFile, (tapeBp * BLOCKSIZE * WORDLEN) + ((tapeBp + 1) * WORDLEN)) || \
				fread((void *) &(reg[DATA1]), WORDLEN, 1, tapeFile) != 1 || reg[DATA1] > TAPEEOB)
			{
				sprintf(strbuf, "Error reading tape %u file : %s", devNum, strerror(errno));
				Panic(strbuf);
			}
			// else read operation is successful:
			if (isWorking)
			{
				if (bus->DMATransfer(tapeBlk, reg[DATA0], TRUE))
				{
					// DMA transfer error
					reg[STATUS] = DMAERR;
					sprintf(statStr, "DMA error reading block %u : waiting for ACK", tapeBp);
				}
				else
				{
					sprintf(statStr, "Block %u read : waiting for ACK", tapeBp);
					reg[STATUS] = READY;
				}
			}
			else
			{
				// no operation & error simulation
				sprintf(statStr, "Error reading block %u : waiting for ACK", tapeBp);
				reg[STATUS] = READERR;
			}
			// tape block counter is however incremented
			tapeBp++;
			break;

		case BACKBLK:
			// a BACKBLK is always successful (isWorking status does not matter)
			tapeBp--;
			if (tapeBp == 0)
				reg[DATA1] = TAPESTART;
			else
				// read previous block for terminator value 
				if (tapeBlk->ReadBlock(tapeFile, ((tapeBp - 1) * BLOCKSIZE * WORDLEN) + (tapeBp * WORDLEN)) || \
				fread((void *) &(reg[DATA1]), WORDLEN, 1, tapeFile) != 1 || reg[DATA1] > TAPEEOB)
			{
				sprintf(strbuf, "Error reading tape %u file : %s", devNum, strerror(errno));
				Panic(strbuf);
			}
			// else operation is successful: 
			reg[STATUS] = READY;
			sprintf(statStr, "Rewound back to block %u : waiting for ACK", tapeBp);
			break;

		default:
			Panic("Unknown operation in TapeDevice::CompleteDevOp()");
			break;
	}
	statChanged = TRUE;
	complTime = NULL;
	bus->IntReq(intL, devNum);
	// here Reg[DATA1] too is changed, but there is only one return value
	// however, if area is traced/suspected fully the result does not change
	return(STATUS);
}


/****************************************************************************/
/* Definitions strictly local to the module.                                */
/****************************************************************************/

// This function decodes device STATUS field and tells if previous operation
// has been successful or not
HIDDEN const char * isSuccess(unsigned int devType, Word regVal)
{
	const char * result = NULL;
	
	switch (devType)
	{
		case PRNTDEV:
		case DISKDEV:
		case TAPEDEV:
			if (regVal == READY)
				result = opResult[TRUE];
			else
				result = opResult[FALSE];
			break;
		
		case TERMDEV:
			if (regVal == READY || regVal == RECVD || regVal == TRANSMD)
				result = opResult[TRUE];
			else
				result = opResult[FALSE];
			break;
		
		default:
			Panic("Unknown device in device module::isSuccess()");
			break;
	}
	return(result);
}	
